import tifffile
from bactfit.fit import bactfit


if __name__ == "__main__":

    mask = tifffile.imread("mask.tif")

    bf = bactfit()

    bf.fit_cells(mask,
        fit=False,
        parallel=False)




import numpy as np
import traceback

import pandas as pd
from shapely.geometry import Point, LineString, Polygon
from shapely.strtree import STRtree
from bactfit.utils import resize_line, rotate_linestring, fit_poly, get_vertical, get_polygon_midline
import matplotlib.pyplot as plt
import math
from shapely.affinity import rotate, translate
import cv2

def find_centerline(midline, radius, smooth=True):

    centerline = None

    try:
        def extract_end_points(line, num_points=2):
            coords = list(line.coords)
            if len(coords) < num_points:
                raise ValueError("The LineString does not have enough points.")
            start_points = coords[:num_points]
            end_points = coords[-num_points:]
            return start_points, end_points

        def extend_away(points, distance, ):
            if len(points) < 2:
                raise ValueError("At least two points are required to determine the direction for extension.")

            p1 = Point(points[0])
            p2 = Point(points[1])

            dx = p2.x - p1.x
            dy = p2.y - p1.y
            length = np.hypot(dx, dy)
            factor = distance / length

            # Extend p1 away from p2
            extended_x1 = p1.x - factor * dx
            extended_y1 = p1.y - factor * dy

            # Similarly for the other end
            p3 = Point(points[-1])
            p4 = Point(points[-2])

            dx_end = p4.x - p3.x
            dy_end = p4.y - p3.y
            length_end = np.hypot(dx_end, dy_end)
            factor_end = distance / length_end

            # Extend p3 away from p4
            extended_x2 = p3.x - factor_end * dx_end
            extended_y2 = p3.y - factor_end * dy_end

            return (extended_x1, extended_y1), (extended_x2, extended_y2)

        model = midline.buffer(width)

        centerline = resize_line(midline, 1000)  # High resolution with 1000 points

        start_points, end_points = extract_end_points(centerline)

        extension_distance = radius * 3

        extended_start = extend_away(start_points, extension_distance)
        extended_end = extend_away(end_points, extension_distance)

        extended_start_line = LineString([start_points[0], extended_start[0]])
        extended_end_line = LineString([end_points[-1], extended_end[1]])

        outline = LineString(model.exterior.coords)
        intersections_start = outline.intersection(extended_start_line).coords[0]
        intersections_end = outline.intersection(extended_end_line).coords[0]

        centerline_coords = np.array(centerline.coords)
        centerline_coords = np.insert(centerline_coords, 0, intersections_start, axis=0)
        centerline_coords = np.append(centerline_coords, [intersections_end], axis=0)
        centerline = LineString(centerline_coords)

        if smooth:
            vertical = get_vertical(model)

            if vertical:
                centerline = rotate_linestring(centerline, angle=90)

            centerline_coords = np.array(centerline.coords)
            constraining_points = [centerline_coords[0], centerline_coords[-1]]

            centerline_coords, _ = fit_poly(centerline_coords, degree=[1, 2, 3], maxiter=100, constraining_points=constraining_points, constrained=True)

            centerline = LineString(centerline_coords)

            if vertical:
                centerline = rotate_linestring(centerline, angle=-90)

    except:
        print(traceback.format_exc())
        pass

    return centerline



def split_linestring(linestring, num_segments):
    points = [linestring.interpolate(float(i) / num_segments, normalized=True) for i in range(num_segments + 1)]
    return [LineString([points[i], points[i + 1]]) for i in range(num_segments)]


def calculate_angle2(segment: LineString, point: Point, centroid: Point) -> float:
    """
    Calculate the angle between the point and the center of the segment, using the center of the source midline.
    """
    
    line_start = segment.coords[0]
    line_end = segment.coords[1]
    
    dx = line_end[0] - line_start[0]
    dy = line_end[1] - line_start[1]
    segment_angle = math.atan2(dy, dx)
    point_angle = math.atan2(point.y - line_start[1], point.x - line_start[0])
    angle = point_angle - segment_angle
    
    return math.degrees(angle)

def calculate_new_point2(segment: LineString, centroid: Point,
                         distance: float, angle: float):
    """
    Calculate new point coordinates in the target segment given the distance and angle.
    """

    line_start = segment.coords[0]
    line_end = segment.coords[1]
    
    dx = line_end[0] - line_start[0]
    dy = line_end[1] - line_start[1]
    segment_angle = math.atan2(dy, dx)
    
    angle_rad = math.radians(angle + math.degrees(segment_angle))
    dx = distance * np.cos(angle_rad)
    dy = distance * np.sin(angle_rad)
    
    new_x = line_start[0] + dx
    new_y = line_start[1] + dy
    
    return new_x, new_y




def calculate_angle(segment, point):
    # Calculate the angle between the segment and the point
    line_start = segment.coords[0]
    line_end = segment.coords[1]
    dx = line_end[0] - line_start[0]
    dy = line_end[1] - line_start[1]
    segment_angle = math.atan2(dy, dx)
    point_angle = math.atan2(point.y - line_start[1], point.x - line_start[0])
    angle = point_angle - segment_angle
    return math.degrees(angle)

def calculate_new_point(segment, distance, angle):
     line_start = segment.coords[0]
     line_end = segment.coords[1]
     segment_angle = math.atan2(line_end[1] - line_start[1], line_end[0] - line_start[0])
     new_angle = segment_angle + angle
     new_x = line_start[0] + distance * math.cos(new_angle)
     new_y = line_start[1] + distance * math.sin(new_angle)
     return [new_x, new_y]

def reflect_loc_horizontally(loc, centroid):

    center_y = centroid[1]
    centroid_distance = loc.y - center_y

    if centroid_distance < 0:
        loc.y = loc.y + abs(centroid_distance)*2
    else:
        loc.y = loc.y - abs(centroid_distance)*2

    return loc

def reflect_loc_vertically(loc, centroid):

    center_x = centroid[0]
    centroid_distance = loc.x - center_x

    if centroid_distance < 0:
        loc.x = loc.x + abs(centroid_distance)*2
    else:
        loc.x = loc.x - abs(centroid_distance)*2

    return loc


def remove_locs_outside_cell(locs, polygon):

    try:

        if type(locs) != np.recarray:
            return None

        polygon_locs = None

        coords = np.stack([locs["x"], locs["y"]], axis=1)
        points = [Point(coord) for coord in coords]
        spatial_index = STRtree(points)

        possible_points = spatial_index.query(polygon)

        polygon_point_indices = []

        for point_index in possible_points:
            point = points[point_index]

            if polygon.contains(point):
                polygon_point_indices.append(point_index)

        if len(polygon_point_indices) > 0:
            polygon_locs = locs[polygon_point_indices]

    except:
        print(traceback.format_exc())
        polygon_locs = None

    return polygon_locs

def compute_vectors(segments):
    unit_vectors = []
    perpendicular_vectors = []
    for segment in segments:
        segment_start = np.array(segment.coords[0])
        segment_end = np.array(segment.coords[1])
        segment_vector = segment_end - segment_start
        segment_length = np.linalg.norm(segment_vector)
        unit_vector = segment_vector / segment_length
        perpendicular_vector = np.array([-unit_vector[1], unit_vector[0]])
        unit_vectors.append(unit_vector)
        perpendicular_vectors.append(perpendicular_vector)
    return unit_vectors, perpendicular_vectors


def cell_coordinate_transformation(cell, target_cell,
        method = "angular", n_segments=1000, progress_list = [], shape_measurements=True):

    if method == "angular":

        cell = angular_coordinate_transformation(cell, target_cell,
            n_segments, progress_list, shape_measurements)

    elif method == "perpendicular":

        cell = perpendicular_coordinate_transformation(cell, target_cell,
            n_segments, progress_list, shape_measurements)

    return cell


def plot_cell(polygon=None, locs=None, midline=None, title=None):

    try:

        if type(polygon) == Polygon:
            polygon_coords = np.array(polygon.exterior.coords)
            plt.plot(*polygon_coords.T, color='black')
        if type(locs) == np.recarray:
            coords = np.stack([locs["x"], locs["y"]], axis=1)
            plt.scatter(*coords.T, color='red', s=1)
        if type(midline) == LineString:
            plt.plot(*np.array(midline.coords).T, color='blue')
        if title:
            plt.title(title)

        plt.show()

    except:
        print(traceback.format_exc())
        pass

def perpendicular_bisector(segment, radius):
    mid_point = segment.interpolate(0.5, normalized=True)
    start, end = segment.coords
    dx, dy = end[0] - start[0], end[1] - start[1]
    bisector = LineString([(-dy, dx), (dy, -dx)])  # Create a perpendicular line
    bisector = translate(bisector, mid_point.x, mid_point.y)  # Translate to midpoint
    scale_factor = radius / bisector.length / 2  # Scale factor to get the desired length on each side
    bisector = LineString([mid_point, (mid_point.x - scale_factor * dy, mid_point.y + scale_factor * dx),
                           (mid_point.x + scale_factor * dy, mid_point.y - scale_factor * dx)])
    return bisector

def build_strtree_index(bisectors):
    return STRtree(bisectors)

def find_closest_bisector_with_strtree(strtree, bisectors, point):
    nearest_geom = strtree.nearest(point)
    closest_index = bisectors.index(nearest_geom)
    min_distance = point.distance(nearest_geom)
    return closest_index, min_distance


def compute_vectors(segments):
    unit_vectors = []
    perpendicular_vectors = []
    for segment in segments:
        segment_start = np.array(segment.coords[0])
        segment_end = np.array(segment.coords[1])
        segment_vector = segment_end - segment_start
        segment_length = np.linalg.norm(segment_vector)
        unit_vector = segment_vector / segment_length
        perpendicular_vector = np.array([-unit_vector[1], unit_vector[0]])
        unit_vectors.append(unit_vector)
        perpendicular_vectors.append(perpendicular_vector)
    return unit_vectors, perpendicular_vectors


def perpendicular_coordinate_transformation(cell, target_cell,
        n_segments=1000, progress_list = [], shape_measurements=True):

    transformed_locs = []

    try:

        cell.locs = remove_locs_outside_cell(cell.locs,
            cell.cell_polygon)

        locs = cell.locs

        if type(locs) != np.recarray:
            return cell

        if len(locs) == 0:
            return cell

        source_polygon = cell.cell_polygon
        source_midline = cell.cell_midline
        source_radius = cell.cell_radius

        if cell.cell_centerline is None:
            cell.cell_centerline = find_centerline(source_midline, source_radius)
            source_centerline = cell.cell_centerline

        target_radius = target_cell.cell_radius
        target_centerline = target_cell.cell_centerline
        target_polygon = target_cell.cell_polygon

        source_segments = split_linestring(source_centerline, n_segments)
        target_segments = split_linestring(target_centerline, n_segments)

        # Precompute vectors for source and target segments
        source_unit_vectors, source_perpendicular_vectors = compute_vectors(source_segments)
        target_unit_vectors, target_perpendicular_vectors = compute_vectors(target_segments)

        # Create STRtree for segments
        tree = STRtree(source_segments)

        for loc in locs:

            try:

                point = Point(loc["x"], loc["y"])

                # Find the nearest segment to each point
                closest_segment_index = tree.nearest(point)

                nearest_segment = source_segments[closest_segment_index]

                segment_start = np.array(nearest_segment.coords[0])
                segment_vector = source_unit_vectors[closest_segment_index]
                point_vector = np.array([loc["x"], loc["y"]]) - segment_start

                # Compute the signed distance from the point to the nearest segment
                distance = point.distance(nearest_segment)
                distance_sign = np.sign(np.cross(segment_vector, point_vector))

                signed_distance = distance*distance_sign
                # Compute the new distance in the target coordinate system
                new_distance = target_radius * (signed_distance / source_radius)

                # Calculate the new coordinates by moving perpendicular to the target segment at the new distance
                segment_start_target = np.array(target_segments[closest_segment_index].coords[0])
                perp_vector = target_perpendicular_vectors[closest_segment_index]
                new_point_coords = segment_start_target + perp_vector * new_distance

                if target_polygon.contains(Point(new_point_coords)):

                    tloc = loc.copy()
                    tloc["x"] = new_point_coords[0]
                    tloc["y"] = new_point_coords[1]

                    transformed_locs.append(tloc)
            except:
                pass

            progress_list.append(1)

        if len(transformed_locs) > 0:
            transformed_locs = np.hstack(transformed_locs).view(np.recarray).copy()

            cell.locs = transformed_locs
            cell.cell_polygon = target_polygon

            if shape_measurements:
                cell = compute_shape_measurements(cell)

        else:
            cell.locs = None

    except:
        print(traceback.format_exc())
        pass

    return cell



def get_pixel_values(image, points, upscale=1):
    
    image = cv2.resize(image, None, fx=upscale, fy=upscale, 
                        interpolation=cv2.INTER_LINEAR)
    
    points = [(point.x, point.y) for point in points]
    
    pixel_values = []
    
    for point in points:
        x, y = point
        # Get the pixel value at the given point
        pixel_values.append(image[int(y), int(x)])
        
    return pixel_values


def scale_polygon(polygon, scale):
    scaled_coords = [(x * scale, y * scale) for x, y in polygon.exterior.coords]
    return Polygon(scaled_coords)

def scale_midline(midline, scale):
    scaled_coords = [(x * scale, y * scale) for x, y in midline.coords]
    return LineString(scaled_coords)

def angular_pixel_transformation(cell, target_cell, upscale = 2, n_segments=1000):
    
    try:
        
        
        if cell.data == {}:
            return None
        
        if cell.cell_polygon == None:
            return None
        
        if cell.cell_midline == None:
            midline, radius = get_polygon_midline(cell.cell_polygon)
            cell.cell_midline = midline
            cell.cell_radius = radius
            
        if cell.cell_midline == None:
            return None
            
        source_polygon = cell.get_image_polygon()
        source_midline = cell.get_image_midline()
        source_polygon = scale_polygon(source_polygon, upscale)
        source_midline = scale_midline(source_midline, upscale)
        source_centroid = source_midline.centroid
        
        source_radius = cell.cell_radius
        target_radius = target_cell.cell_radius
        target_image = target_cell.get_image()


        target_midline = target_cell.cell_midline
        target_polygon = target_cell.cell_polygon
        target_centroid = target_polygon.centroid
        
        source_segments = split_linestring(source_midline, n_segments)
        target_segments = split_linestring(target_midline, n_segments)
        
        # Create STRtree for segments
        tree = STRtree(source_segments)
        
        # Calculate the centroid of the midline
        centroid = source_polygon.centroid
        cx, cy = int(centroid.x), int(centroid.y)
        
        # Get the bounding box of the polygon
        minx, miny, maxx, maxy = map(int, source_polygon.bounds)
        
        # Generate all integer points within the bounding box
        points = [Point((x, y)) for x in range(minx, maxx + 1) for y in range(miny, maxy + 1)]
        
        # Filter points that are inside the polygon
        points = [point for point in points if source_polygon.contains(point) or 
                  source_polygon.touches(point)]
        
        # points_coords = np.array([point.coords for point in points])
        # plt.scatter(*points_coords.T)
        # plt.show()
        
        # image = cell.data["NR"]
        # image = cv2.resize(image, None, fx=upscale, fy=upscale, 
        #                     interpolation=cv2.INTER_LINEAR)
        
        pixel_dict = {channel: get_pixel_values(image, points, upscale) for channel, image in cell.data.items()}
        
        transformed_data = {channel:target_image.copy() for channel in cell.data.keys()}
        
        n_transformed = 0
        
        polygon_coords = np.array(target_polygon.exterior.coords)
        midline_coords = np.array(target_midline.coords)
        
        for point_index, point in enumerate(points):

            try:
                
                if point_index <= 10000:

                    closest_segment_index = tree.nearest(point)
                    nearest_segment = source_segments[closest_segment_index]
    
                    source_distance = point.distance(nearest_segment)
                    angle = calculate_angle2(nearest_segment, point, source_centroid)
                    
                    if source_distance > source_radius:
                        source_distance = source_radius
                        
                    target_distance = target_radius * (source_distance / source_radius)
                    
                    # Use the corresponding target segment
                    target_segment = target_segments[closest_segment_index]
    
                    # Calculate the new point in the target segment
                    new_point_coords = calculate_new_point2(target_segment, target_centroid, target_distance, angle)
                    new_point = Point(new_point_coords)
                    
                if target_polygon.contains(new_point) or target_polygon.touches(new_point):
                    
                    n_transformed +=1
                    
                    plt.scatter(new_point_coords[0], new_point_coords[1])
                
                    for channel, image in transformed_data.items():
                        
                        pixel_value = pixel_dict[channel][point_index]
                        iX, iY = new_point_coords
                        
                        image[int(iY),int(iX)] = pixel_value
                        transformed_data[channel] = image
                        
            except:
                print(traceback.format_exc())
                
        plt.show()
                


        # if transformed_data != {}:
        #     target_cell.data = transformed_data

    except:
        print(traceback.format_exc())
        
    return target_cell


def check_point_inside_polygon(polygon, point):
    
    if isinstance(polygon, np.ndarray):
        polygon = Polygon(polygon)

    if type(point) in [np.ndarray, list, tuple]:
        point = Point(point)
        
    if polygon.contains(point) or polygon.touches(point):
        return True
    else:
        return False


def plot_cell(cell):

    try:

        if cell.cell_polygon is not None:
            polygon_coords = np.array(cell.cell_polygon.exterior.coords)
            plt.plot(*polygon_coords.T, color='black')

        if cell.cell_midline is not None:
            midline_coords = np.array(cell.cell_midline.coords)
            plt.plot(*midline_coords.T, color='blue')

        if cell.locs is not None:
            loc_coords = np.stack([cell.locs["x"], cell.locs["y"]], axis=1)
            plt.scatter(*loc_coords.T, color='red', s=50)

        plt.show()

    except:
        print(traceback.format_exc())
        pass


def angular_coordinate_transformation(cell, target_cell,
        n_segments=1000, shape_measurements=True, progress_list = []):


    try:

        cell.locs = remove_locs_outside_cell(cell.locs,
            cell.cell_polygon)

        locs = cell.locs

        if type(locs) != np.recarray:
            return cell

        if len(locs) == 0:
            return cell

        source_polygon = cell.cell_polygon
        source_midline = cell.cell_midline

        if source_polygon == None:
            return None
        if source_midline == None:
            return None

        source_centroid = source_midline.centroid

        source_radius = cell.cell_radius
        target_radius = target_cell.cell_radius

        target_midline = target_cell.cell_midline
        target_polygon = target_cell.cell_polygon
        target_centroid = target_polygon.centroid

        source_segments = split_linestring(source_midline, n_segments)
        target_segments = split_linestring(target_midline, n_segments)

        # Create STRtree for segments
        tree = STRtree(source_segments)

        transformed_locs = []

        for loc in locs:

            try:

                point = Point(loc["x"], loc["y"])

                closest_segment_index = tree.nearest(point)
                nearest_segment = source_segments[closest_segment_index]

                source_distance = point.distance(nearest_segment)
                angle2 = calculate_angle2(nearest_segment, point, source_centroid)

                target_distance = target_radius * (source_distance / source_radius)
                
                # Use the corresponding target segment
                target_segment = target_segments[closest_segment_index]

                # Calculate the new point in the target segment
                new_point_coords = calculate_new_point2(target_segment, target_centroid, 
                                                        target_distance, angle2)

                if len(new_point_coords) == 0:
                    continue

                if check_point_inside_polygon(target_polygon, new_point_coords):

                    tloc = loc.copy()
                    tloc["x"] = new_point_coords[0]
                    tloc["y"] = new_point_coords[1]

                    transformed_locs.append(tloc)

            except:
                print(traceback.format_exc())
                pass

        cell.cell_polygon = target_polygon
        cell.cell_midline = target_midline

        if len(transformed_locs) > 0:

            transformed_locs = np.hstack(transformed_locs).view(np.recarray).copy()

            cell.locs = transformed_locs

            if shape_measurements:
                cell = compute_shape_measurements(cell)

        else:
            cell.locs = None

    except:
        print(traceback.format_exc())
        cell.locs = None
        pass

    return cell

def compute_shape_measurements(cell):

    try:

        locs = cell.locs
        midline = cell.cell_midline
        polygon = cell.cell_polygon
        pixel_size = cell.pixel_size

        locs = pd.DataFrame(locs)
        locs["pixel_size (nm)"] = pixel_size
        locs["centroid_distance (nm)"] = None
        locs["membrane_distance (nm)"] = None
        locs["midline_distance (nm)"] = None
        locs["cell_pole_distance (nm)"] = None

        if len(locs) == 0:
            return None

        coords = np.array(locs[["x", "y"]].values)

        try:

            polygon_centroid = polygon.centroid.coords[0]

            centroid_distance = [Point(coord).distance(Point(polygon_centroid)) for coord in coords]

            if len(centroid_distance) == len(coords):
                centroid_distance = np.array(centroid_distance)
                centroid_distance = centroid_distance * pixel_size
                centroid_distance = centroid_distance.tolist()
                locs["centroid_distance (nm)"] = centroid_distance

                print(centroid_distance)

        except:
            pass

        try:

            polygon_coords = np.array(polygon.exterior.coords)
            polygon_outline = LineString(polygon_coords)

            membrane_distance = [Point(coord).distance(polygon_outline) for coord in coords]

            if len(membrane_distance) == len(coords):
                membrane_distance = np.array(membrane_distance)
                membrane_distance = membrane_distance * pixel_size
                membrane_distance = membrane_distance.tolist()
                locs["membrane_distance (nm)"] = membrane_distance

        except:
            pass

        try:

            midline_distance = [Point(coord).distance(midline) for coord in coords]

            if len(midline_distance) == len(coords):
                midline_distance = np.array(midline_distance)
                midline_distance = midline_distance * pixel_size
                midline_distance = midline_distance.tolist()
                locs["midline_distance (nm)"] = midline_distance

        except:
            pass

        try:
            midline_coords = np.array(midline.coords)
            cell_poles = [Point(midline_coords[0]), Point(midline_coords[-1])]

            cell_pole_distance = [min([Point(coord).distance(pole) for pole in cell_poles]) for coord in coords]

            if len(cell_pole_distance) == len(coords):
                cell_pole_distance = np.array(cell_pole_distance)
                cell_pole_distance = cell_pole_distance * pixel_size
                cell_pole_distance = cell_pole_distance.tolist()
                locs["cell_pole_distance (nm)"] = cell_pole_distance
        except:
            pass

        if len(locs) > 0:
            cell.locs = locs.to_records(index=False)

    except:
        print(traceback.format_exc())
        pass

    return cell





import random
import string
import time
import copy
import numpy as np
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed, ProcessPoolExecutor
from multiprocessing import Manager, Event
from functools import partial
from scipy.spatial import distance
import os
import pickle
from shapely.geometry import Polygon, LineString, Point
from shapely.strtree import STRtree
import pandas as pd
import traceback
import matplotlib.pyplot as plt
from picasso.render import render
import warnings
from shapely.affinity import translate

from bactfit.fit import bactfit
from bactfit.postprocess import (cell_coordinate_transformation,
    reflect_loc_horizontally, reflect_loc_vertically)
from bactfit.utils import resize_line, get_vertical
import h5py
import cv2
from matplotlib.colors import ListedColormap
from io import BytesIO
from picasso.render import render



class ModelCell(object):

    def __init__(self, length = 10, radius = 5, margin = 1):

        from bactfit.__init__ import __version__ as bactfit_version
        self.bactfit_version = bactfit_version
        self.type = "ModelCell"

        self.cell_polygon = None
        self.cell_midline = None
        self.cell_centerline = None
        self.cell_radius = radius
        self.cell_length = length
        self.margin = margin

        self.create_model_cell()

    def create_model_cell(self):

        x0 = y0 = self.cell_radius + self.margin

        # Define the coordinates of the line
        midline_x_coords = [x0, x0 + self.cell_length]
        midline_y_coords = [y0, y0]
        midline_coords = list(zip(midline_x_coords, midline_y_coords))
        self.cell_midline = LineString(midline_coords)
        self.cell_polygon = self.cell_midline.buffer(self.cell_radius)

        y0 = self.cell_radius + self.margin
        x0 = self.margin
        centerline_x_coords = [x0, x0 + self.cell_length + (self.cell_radius * 2)]
        centerline_y_coords = [y0, y0]
        centerline_coords = list(zip(centerline_x_coords, centerline_y_coords))
        centerline_coords = np.array(centerline_coords)

        self.cell_centerline = LineString(centerline_coords)
        self.cell_centerline = resize_line(self.cell_centerline, 100)

    def get_image(self, dtype = np.uint16):
        
        width = self.cell_length + ((self.cell_radius+ self.margin) * 2)
        height = self.cell_radius*2 + (self.margin * 2)
        
        image = np.zeros((height,width), dtype=np.uint16)
        
        return image
        

        


class Cell(object):

    def __init__(self, cell_data = None):

        from bactfit.__init__ import __version__ as bactfit_version
        self.bactfit_version = bactfit_version
        self.type = "Cell"

        self.cell_polygon = None
        self.cell_centre = None
        self.bbox = None
        self.height = None
        self.cell_radius = None
        self.vertical = None
        self.name = None
        self.transformed = False

        self.data = {}
        self.locs = []

        #fit data
        self.cell_midline = None
        self.cell_centerline = None
        self.cell_poles = None
        self.cell_index = None
        self.polynomial_params = None
        self.fit_error = None
        self.pixel_size = None

        if cell_data is not None:

            if type(cell_data) == dict:

                for key in cell_data.keys():
                    if key == "radius":
                        setattr(self, "cell_radius", cell_data[key])
                    elif key == "length":
                        setattr(self, "cell_length", cell_data[key])
                    else:
                        setattr(self, key, cell_data[key])

            elif type(cell_data) == np.ndarray:

                cell_mask = cell_data

                if len(cell_mask.shape) != 2:
                    print("Cell data must be a 2D numpy array")
                unique_ids = np.unique(cell_mask)
                if len(unique_ids) != 2:
                    print("Cell data must be a binary mask")
                self.import_mask(cell_mask)

            else:
                print("Cell initialisation must be a dictionary or a 2D numpy array")

        if self.name == None:
            self.generate_name()
            
        if hasattr(self, "midline_coords") and self.cell_midline is None:
            self.cell_midline = LineString(self.midline_coords)
            self.cell_poles = [self.cell_midline.coords[0], self.cell_midline.coords[-1]]

        if hasattr(self, "poly_params") and self.polynomial_params is None:
            self.polynomial_params = self.poly_params

        if self.cell_midline is not None and self.cell_radius is not None:
            self.cell_polygon = self.cell_midline.buffer(self.cell_radius)

        if self.vertical is None and self.cell_polygon is not None:
            self.vertical = get_vertical(self.cell_polygon)

    def __getstate__(self):
        # Return the state as a dictionary, omitting non-picklable attributes
        state = self.__dict__.copy()
        # Remove attributes that cannot be pickled
        # state.pop('non_picklable_attribute', None)
        return state

    def generate_name(self):
        self.name = ''.join(random.choices(string.ascii_uppercase + string.digits, k=10))
        
    def import_mask(self, cell_mask):

        try:
            contours, _ = cv2.findContours(cell_mask.astype(np.uint8),
                cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)

            if len(contours) > 0:
                contour = contours[0]
                contour = contour.squeeze()

                cell_polygon = Polygon(contour)

                centroid = cell_polygon.centroid
                cell_centre = [centroid.x, centroid.y]

                minx, miny, maxx, maxy = cell_polygon.bounds

                bbox = [minx, miny, maxx, maxy]

                h = maxy - miny
                w = maxx - minx

                if h > w:
                    vertical = True
                else:
                    vertical = False

                self.cell_polygon = cell_polygon
                self.cell_centre = cell_centre
                self.bbox = bbox
                self.height = h
                self.width = w
                self.vertical = vertical
                self.frame_index = 0

        except:
            print(traceback.format_exc())
            return None

    def __setstate__(self, state):
        # Restore the state
        self.__dict__.update(state)

    def remove_locs_outside_cell(self, locs=None):

        if locs is not None:

            if isinstance(locs, pd.DataFrame):
                locs = locs.to_records(index=False)

            self.locs = locs

        if self.locs is None:
            return None

        if len(self.locs) == 0:
            self.locs = None
            return None

        filtered_locs = []

        coords = np.stack([self.locs["x"], self.locs["y"]], axis=1)
        points = [Point(coord) for coord in coords]
        spatial_index = STRtree(points)

        possible_points = spatial_index.query(self.cell_polygon)

        polygon_point_indices = []

        for point_index in possible_points:
            point = points[point_index]

            if self.cell_polygon.contains(point):
                polygon_point_indices.append(point_index)

        if len(polygon_point_indices) > 0:
            polygon_locs = self.locs[polygon_point_indices]

            polygon_locs = pd.DataFrame(polygon_locs)
            polygon_locs = polygon_locs.to_records(index=False)

            filtered_locs.append(polygon_locs)

        if len(filtered_locs) > 0:
            filtered_locs = np.hstack(filtered_locs).view(np.recarray).copy()
            self.locs = filtered_locs
        else:
            self.locs = None

    def transform_cells(self, target_cell=None, locs=None,
            method = "angular", shape_measurements=True):

        if locs is not None:
            self.locs = locs

        if target_cell is not None and self.locs is not None:

            cell = cell_coordinate_transformation(cell=self, target_cell=target_cell,
                method=method, shape_measurements=shape_measurements)

            cell_locs = cell.locs

            if cell_locs is not None:
                self.locs = cell_locs
            else:
                self.locs = None

    def optimise(self, refine_fit = True, fit_mode = "directed_hausdorff",
            min_radius = -1, max_radius = -1, max_error = 5, **kwargs):

        try:
            bf = bactfit(cell=self, refine_fit=refine_fit, fit_mode=fit_mode,
                min_radius=min_radius, max_radius=max_radius, max_error=max_error, **kwargs)

            self = bf.fit()

        except:
            print(traceback.format_exc())
            return None

        return self

    def plot(self):

        if self.cell_polygon is not None:
            polygon = self.cell_polygon
        else:
            polygon = None

        if self.cell_midline is not None:
            midline = self.cell_midline
        else:
            midline = None

        if self.locs is not None:
            locs = self.locs
        else:
            locs = None

        if polygon is not None:
            ploygon_coords = np.array(polygon.exterior.coords)
            plt.plot(*ploygon_coords.T, color="black")
        if midline is not None:
            midline_coords = np.array(midline.coords)
            plt.plot(*midline_coords.T, color="red")
        if locs is not None:
            plt.scatter(locs["x"], locs["y"], color="blue")
        plt.show()

    def add_image(self, image, image_name):

        try:

            if type(image) != np.ndarray:
                print("Image is not a numpy array")
                return None
            if len(image.shape) != 2:
                print("Image is not 2D")
                return None

            if self.cell_polygon is not None:

                cropped_image = self.crop_image(image)

                if cropped_image is None:
                    return None

                self.data[image_name] = cropped_image
        except:
            print(traceback.format_exc())

    def crop_image(self, image, buffer=10):

        cropped_image = None

        try:
            image_shape = image.shape

            minx, miny, maxx, maxy = self.cell_polygon.bounds

            minx = int(minx) - buffer
            miny = int(miny) - buffer
            maxx = int(maxx) + buffer
            maxy = int(maxy) + buffer

            if minx < 0:
                minx = 0
            if miny < 0:
                miny = 0
            if maxx > image_shape[1]:
                maxx = image_shape[1]
            if maxy > image_shape[0]:
                maxy = image_shape[0]

            cropped_image = image[miny:maxy, minx:maxx]

            self.crop_bounds = [minx, miny, maxx, maxy]

        except:
            print(traceback.format_exc())

        return cropped_image

    def get_image(self, image_name):

        if image_name not in self.data.keys():
            return None

        image = self.data[image_name]

        return image

    def get_image_polygon(self):

        if self.cell_polygon is None:
            return None
        if self.crop_bounds is None:
            return None

        minx, miny, maxx, maxy = self.crop_bounds

        polygon = self.cell_polygon

        polygon = translate(polygon, xoff=-minx, yoff=-miny)

        return polygon
    
    def get_image_midline(self):
        
        if self.cell_midline is None:
            return None
        if self.crop_bounds is None:
            return None

        minx, miny, maxx, maxy = self.crop_bounds

        midline = self.cell_midline

        midline = translate(midline, xoff=-minx, yoff=-miny)

        return midline
        
    def get_image_mask(self):

        if self.cell_polygon is None:
            return None
        if self.crop_bounds is None:
            return None

        minx, miny, maxx, maxy = self.crop_bounds

        polygon = self.cell_polygon

        polygon = translate(polygon, xoff=-minx, yoff=-miny)

        polygon_coords = np.array(polygon.exterior.coords)
        mask = np.zeros((maxy-miny, maxx-minx))
        mask = cv2.fillPoly(mask, [polygon_coords.astype(int)], 255)
        mask = mask.astype(bool)

        return mask



    

        

class CellList(object):

    def __init__(self, cell_list, verbose=False):

        from bactfit.__init__ import __version__ as bactfit_version
        self.bactfit_version = bactfit_version
        self.type = "CellList"

        if isinstance(cell_list[0], Cell):
            self.data = cell_list

        elif isinstance(cell_list[0], CellList):
            self.data = []
            for cell_list in cell_list:
                self.data.extend(cell_list.data)

        self.cell_names = []
        self.verbose = verbose

        self.assign_cell_indices()
        self.assign_cell_names()

    def assign_cell_indices(self, reindex=False):

        if reindex == False:
            self.cell_indices = [cell.cell_index for cell in self.data
                                 if cell.cell_index is not None]
        else:
            self.cell_indices = []

        if len(self.cell_indices) == 0:

            for i, cell in enumerate(self.data):
                cell.cell_index = i
                self.cell_indices.append(i)

    def assign_cell_names(self, rename=False):

        if rename == False:
            self.cell_names = [cell.name for cell in self.data if cell.name is not None]
        else:
            self.cell_names = []

        if len(self.cell_names) == 0:

            for cell in self.data:
                cell.name = ''.join(random.choices(string.ascii_uppercase + string.digits, k=10))
                self.cell_names.append(cell.name)

    @staticmethod
    def compute_optimisation(jobs, progress_list):

        optimised_cells = []

        for job_index, job in enumerate(jobs):

            cell, optimisation_args = job

            try:
                cell = cell.optimise(**optimisation_args)

                if cell is not None:
                    optimised_cells.append(cell)
                    progress_list.append(1)

            except:
                print(traceback.format_exc())
                pass

        return optimised_cells

    def optimise(self, refine_fit=True, fit_mode="directed_hausdorff",
            min_radius = -1, max_radius = -1, max_workers=None, progress_callback=None,
            silence_tqdm=False, parallel=True, max_error=5):

        try:

            if len(self.data) == 0:
                return None

            optimisation_args = {"refine_fit": refine_fit, "fit_mode": fit_mode,
                "min_radius": min_radius, "max_radius": max_radius,
                "max_error": max_error}

            jobs = [(cell, optimisation_args) for cell in self.data]

            n_jobs = len(jobs)
            n_optimised = 0

            if isinstance(jobs[0][0], Cell) and parallel is True:
                executor = ProcessPoolExecutor()
                n_cpus = os.cpu_count()
                if isinstance(max_workers, int):
                    n_cpus = min(n_cpus, max_workers)
                compute_jobs = np.array_split(jobs, n_cpus)
            else:
                executor = ThreadPoolExecutor()
                n_cpus = 1
                compute_jobs = [[job] for job in jobs]

            print(f"Optimising {n_jobs} cells using {n_cpus} CPU(s) cores")

            optimised_cells = []

            with Manager() as manager:

                progress_list = manager.list()

                with executor:

                    futures = [executor.submit(CellList.compute_optimisation,
                        job, progress_list) for job in compute_jobs]

                    self.compute_progress(futures, n_jobs,
                        progress_list, progress_callback,
                        silence_tqdm=silence_tqdm, tqdm_dec="Optimising Cells")

                    for future in as_completed(futures):
                        try:
                            results = future.result()

                            if results is not None:
                                if len(results) > 0:
                                    optimised_cells.extend(results)
                                    n_optimised += len(results)

                        except Exception as e:
                            print(f"Error: {e}")
                            traceback.print_exc()

            optimised_cells = [cell for cell in optimised_cells if cell is not None]
            self.data = optimised_cells

            print(f"Optimised {n_optimised}/{n_jobs} cells")

            if len(self.data) > 0:
                self.assign_cell_indices(reindex=True)
                self.assign_cell_names()

            error_list = [cell.fit_error for cell in self.data]
            error_list = [error for error in error_list if error is not None]
            print(f"Max fit error: {max(error_list)}")

        except:
            print(traceback.format_exc())
            return None

        return self

    def get_cell_polygons(self, ndim=2, flipxy=False):

        polygons = []
        poly_params = []
        cell_poles = []
        midlines = []
        cell_radii = []
        names = []
        frame_indices = []

        for cell in self.data:
            if hasattr(cell, "cell_polygon"):

                try:

                    cell_polygon = cell.cell_polygon
                    cell_radius = cell.cell_radius
                    cell_midline = cell.cell_midline
                    params = cell.polynomial_params
                    poles = cell.cell_poles

                    if hasattr(cell, "frame_index"):
                        frame_index = cell.frame_index
                    else:
                        frame_index = 0

                    if cell_polygon is not None:

                        name = cell.name
                        cell_polygon = cell_polygon.simplify(0.2)
                        seg = np.array(cell_polygon.exterior.coords)

                        midline = resize_line(cell_midline, 6)
                        midline = np.array(midline.coords)

                        seg = seg[1:]

                        if flipxy:
                            seg = np.fliplr(seg)
                            midline = np.fliplr(midline)

                        if ndim == 3:
                            seg = np.vstack([np.ones(len(seg))*frame_index, seg.T]).T
                            midline = np.vstack([np.ones(len(midline))*frame_index, midline.T]).T

                        polygons.append(seg)
                        names.append(name)
                        midlines.append(midline)
                        cell_radii.append(cell_radius)
                        poly_params.append(params)
                        cell_poles.append(poles)
                        frame_indices.append(frame_index)

                except:
                    pass

        data = {"polygons": polygons,
                "midlines": midlines,
                "cell_radii": cell_radii,
                "names": names,
                "poly_params": poly_params,
                "cell_poles": cell_poles,
                "frame_indices": frame_indices,
                }

        return data

    @staticmethod
    def add_localisation_task(cells, locs, remove_outside, progress_list):

        try:

            for cell in cells:
                cell.locs = locs

                if remove_outside is True:
                    cell.remove_locs_outside_cell()

        except:
            print(traceback.format_exc())
            pass

        return cells

    def add_localisations(self, locs, remove_outside = True, parallel=True,
            silence_tqdm=False, progress_callback=None):

        if isinstance(locs, pd.DataFrame):
            locs = locs.to_records(index=False)

        if isinstance(locs, np.recarray) == False:
            print("Localisations must be a numpy recarray or pandas dataframe")
            return None

        locs = np.unique(locs)

        jobs = [cell for cell in self.data]

        n_jobs = len(jobs)

        if parallel is True and len(jobs) > 100:
            executor = ProcessPoolExecutor()
            n_cpus = os.cpu_count()
            compute_jobs = np.array_split(jobs, n_cpus)
        else:
            executor = ThreadPoolExecutor()
            n_cpus = 1
            compute_jobs = [[job] for job in jobs]

        print(f"Adding {len(locs)} localisations to {n_jobs} cells using {n_cpus} CPU(s) cores")

        with Manager() as manager:

            progress_list = manager.list()

            with executor:

                futures = [executor.submit(CellList.add_localisation_task,
                    job, locs, remove_outside, progress_list) for job in compute_jobs]

                self.compute_progress(futures, n_jobs, progress_list, progress_callback,
                    silence_tqdm=silence_tqdm, tqdm_dec="Adding Localisations")

                for future in as_completed(futures):
                    try:
                        results = future.result()

                        if results is not None:
                            if len(results) > 0:
                                for cell in results:
                                    cell_index = cell.cell_index
                                    locs = cell.locs

                                    if type(locs) == np.recarray:
                                        self.data[cell_index] = cell
                                    else:
                                        self.data[cell_index].locs = None

                    except Exception as e:
                        print(f"Error: {e}")
                        traceback.print_exc()

        locs = self.get_locs()

        if locs is None:
            print(f"Added 0 localisations to CellList")

        if len(locs) > 0:
            n_locs = len(locs)
        else:
            n_locs = 0

        print(f"Added {n_locs} localisations to CellList")

    @staticmethod
    def compute_transforms(jobs, compute_list):
        
        transformed_cells = []

        for job_index, job in enumerate(jobs):

            cell, target_cell, method, shape_measurements = job
            cell = cell_coordinate_transformation(cell=cell, target_cell=target_cell,
                method=method, shape_measurements=shape_measurements)

            if cell is not None:
                cell.transformed = True
                transformed_cells.append(cell)

            compute_list.append(1)

        return transformed_cells
        

    def is_optimised(self, max_error = 5):

        optimised = True

        for cell in self.data:
            if cell.fit_error is None:
                optimised = False
            if cell.fit_error > max_error:
                optimised = False

        return optimised


    def compute_progress(self, futures, n_jobs, progress_list,
            progress_callback=None, silence_tqdm=False, tqdm_dec="Optimising Cells"):

        try:
            last_progress = None
            with tqdm(total=n_jobs, desc=tqdm_dec, disable=silence_tqdm) as pbar:
                while True:
                    progress = ((len(progress_list) + 1) / n_jobs)
                    progress = int(progress * 100)

                    if progress != last_progress:
                        pbar.update(progress - (last_progress or 0))
                        last_progress = progress
                        if progress_callback is not None:
                            progress_callback.emit(progress)

                    time.sleep(0.1)  # Sleep for a short duration to reduce CPU usage

                    if all(future.done() for future in futures):
                        if progress_callback is not None:
                            progress_callback.emit(100)
                        pbar.update(n_jobs - pbar.n)
                        break
        except:
            print(traceback.format_exc())
            return

    def transform_cells(self, target_cell=None, method = "angular",
            shape_measurements=True, progress_callback=None, silence_tqdm=False):

        if target_cell is not None:

            cell_with_locs = [cell for cell in self.data if cell.locs is not None]

            jobs = [list([cell,target_cell,
                          method, shape_measurements]) for cell in cell_with_locs if len(cell.locs) > 0]

            n_jobs = len(jobs)
            n_transformed = 0

            if isinstance(jobs[0][0], Cell) and n_jobs > 1:
                executor = ProcessPoolExecutor()
                n_cpus = os.cpu_count()
                compute_jobs = np.array_split(jobs, n_cpus)
            else:
                executor = ThreadPoolExecutor()
                n_cpus = 1
                compute_jobs = [[job] for job in jobs]

            print(f"Transforming localisations within {n_jobs} cells using {n_cpus} CPU(s) cores")

            with Manager() as manager:

                progress_list = manager.list()

                with executor:

                    futures = [executor.submit(CellList.compute_transforms, job, progress_list) for job in compute_jobs]

                    self.compute_progress(futures, n_jobs, progress_list, progress_callback,
                        silence_tqdm=silence_tqdm, tqdm_dec="Transforming Cells")

                    for future in as_completed(futures):

                        try:
                            results = future.result()

                            if results is not None:

                                if len(results) > 0:
                                    for cell in results:
                                        cell_index = cell.cell_index
                                        locs = cell.locs

                                        if type(locs) == np.recarray:
                                            self.data[cell_index] = cell
                                            n_transformed += 1
                                        else:
                                            self.data[cell_index].locs = None

                        except Exception as e:
                            print(f"Error: {e}")
                            traceback.print_exc()

            print(f"Transformed {n_transformed}/{n_jobs} cells")

            for cell in self.data:
                cell.cell_polygon = target_cell.cell_polygon
                cell.cell_midline = target_cell.cell_midline
                cell.cell_radius = target_cell.cell_radius
                cell.remove_locs_outside_cell()

            if len(self.data) > 0:
                self.assign_cell_indices(reindex=True)
                self.assign_cell_names()

        return self

    def get_locs(self, symmetry=False, transformed=None):

        locs = []

        for cell in self.data:
            try:
                cell_locs = cell.locs

                if cell_locs is None:
                    continue
                if len(cell_locs) == 0:
                    continue

                if transformed is not None:
                    if cell.transformed != transformed:
                        continue

                if isinstance(cell_locs, np.recarray):
                    locs.append(cell_locs)

            except:
                continue

        if len(locs) > 0:

            locs = np.hstack(locs).view(np.recarray).copy()
            locs = np.unique(locs)

            if symmetry:
                midline = self.data[0].cell_midline
                centroid_coords = np.array(midline.centroid.coords[0])

                locs = [loc for loc in locs]

                for loc_index in range(len(locs)):
                    loc = locs[loc_index].copy()
                    rloc = reflect_loc_horizontally(loc, centroid_coords)
                    locs.append(rloc)
                for loc_index in range(len(locs)):
                    loc = locs[loc_index].copy()
                    rloc = reflect_loc_vertically(loc, centroid_coords)
                    locs.append(rloc)

                locs = np.hstack(locs).view(np.recarray).copy()

            return locs
        else:
            return None

    def get_cell_lengths(self):

        self.cell_lengths = []

        for cell in self.data:

            try:
                radius = cell.cell_radius
                cell_midline = cell.cell_midline
                midline_length = cell_midline.length
                length = midline_length + (radius * 2)
                pixel_size_nm = cell.pixel_size
                length_um = length * pixel_size_nm / 1000
                cell.cell_length_um = length_um

                self.cell_lengths.append(length_um)
            except:
                print(traceback.format_exc())
                pass

        return self.cell_lengths

    def filter_by_length(self, min_length = 0, max_length = 0):

        if hasattr(self, "cell_lengths") == False:
            self.get_cell_lengths()

        filtered_cells = []

        for i, cell in enumerate(self.data):

            try:
                length_um = cell.cell_length_um

                if length_um >= min_length and length_um <= max_length:
                    filtered_cells.append(cell)
            except:
                pass

        filtered_cells = CellList(filtered_cells)

        return filtered_cells

    def plot_cell_heatmap(self, locs=None, color="red"):

        plot_locs = []

        for cell in self.data:
            locs = cell.locs
            if locs is not None and len(locs) > 0:
                plot_locs.append(locs)

        if len(plot_locs) > 0:
            locs = np.hstack(plot_locs).view(np.recarray).copy()

            # create heatmap
            heatmap, xedges, yedges = np.histogram2d(locs["x"], locs["y"], bins=30, density=False)
            extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
            plt.rcParams["axes.grid"] = False
            plt.imshow(heatmap.T, extent=extent, origin='lower')
            plt.show()

    def plot_cell_render(self, oversampling=10, pixel_size=1, blur_method = "One-Pixel-Blur",
            locs=None, color="red"):

        plot_locs = []

        for cell in self.data:
            locs = cell.locs
            if locs is not None and len(locs) > 0:
                plot_locs.append(locs)

        if len(plot_locs) > 0:
            locs = np.hstack(plot_locs).view(np.recarray).copy()

            print(f"Rendering {len(locs)} localisations")

            xmin, xmax = int(np.min(locs["x"])), int(np.max(locs["x"]))
            ymin, ymax = int(np.min(locs["y"])), int(np.max(locs["y"]))

            h = ymax-ymin
            w = xmax-xmin

            viewport = [(float(0), float(0)), (float(h), float(w))]
            image_shape = (1, int(h), int(w))

            if blur_method == "One-Pixel-Blur":
                blur_method = "smooth"
            elif blur_method == "Global Localisation Precision":
                blur_method = "convolve"
            elif blur_method == "Individual Localisation Precision, iso":
                blur_method = "gaussian_iso"
            elif blur_method == "Individual Localisation Precision":
                blur_method = "gaussian"
            else:
                blur_method = None

            # n_rendered_locs, image = render(locs, viewport=viewport, blur_method=blur_method,
            #     min_blur_width=0, oversampling=oversampling, ang=0, )
            #
            # plt.imshow(image)
            # plt.show()

    def add_image(self, image, image_name):

        for cell in self.data:
            cell.add_image(image, image_name)


    def get_custom_cmap(self, colour = "jet"):

        cmap = plt.get_cmap(colour.lower())
        new_cmap = cmap(np.arange(cmap.N))

        new_cmap[0] = [0, 0, 0, 1]

        new_cmap = ListedColormap(new_cmap)

        return new_cmap

    def plot_heatmap(self, symmetry=False, bins=20, cmap="jet",
            draw_outline=True, show=True, save=False, path=None,dpi=500):

        heatmap = None

        try:
            locs = self.get_locs(symmetry=symmetry, transformed=True)

            if locs is None:
                return None

            if len(locs) == 0:
                return None

            polygon = self.data[0].cell_polygon
            polygon_coords = np.array(polygon.exterior.coords)

            n_cells = len(self.data)
            n_locs = len(locs)

            if symmetry:
                n_locs = int(n_locs/4)

            print(f"Generating Cell heatmap with {n_locs} localisations from {n_cells} cells")

            heatmap, xedges, yedges = np.histogram2d(locs["x"], locs["y"], bins=bins)
            extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]

            cmap = self.get_custom_cmap(colour=cmap)

            plt.rcParams["axes.grid"] = False
            fig, ax = plt.subplots()
            im = ax.imshow(heatmap.T, extent=extent, origin='lower', cmap=cmap)
            if draw_outline:
                ax.plot(*polygon_coords.T, color='white', linewidth=1)
            ax.axis('off')

            from mpl_toolkits.axes_grid1 import make_axes_locatable
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="2%", pad=0.05)
            cbar = plt.colorbar(im, cax=cax)

            # Set the face color of the colorbar to black and the tick color to white
            cax.set_facecolor('black')
            cbar.ax.yaxis.set_tick_params(color='white')
            cbar.outline.set_edgecolor('white')
            plt.setp(plt.getp(cbar.ax.axes, 'yticklabels'), color='white')

            buf = BytesIO()
            plt.savefig(buf, format='png',
                bbox_inches='tight', pad_inches=0.1,
                facecolor='black', dpi=dpi)
            buf.seek(0)
            heatmap = plt.imread(buf)

            plt.close(fig)

            if show:
                plt.imshow(heatmap)
                plt.axis('off')
                plt.show()

            if save:
                try:
                    if path is None:
                        print("Save path must be specified")
                        return None
                    plt.imsave(path, heatmap)
                    print(f"Saved heatmap to {path}")
                except:
                    print(traceback.format_exc())
                    return None

            plt.close()

        except:
            print(traceback.format_exc())
            pass

        return heatmap

    def plot_render(self, symmetry=False, oversampling=10, pixel_size=1,
            blur_method="One-Pixel-Blur", min_blur_width=0.2,
            show=True, save=False, path=None, dpi=500,
            cmap="jet", draw_outline=True):

        image = None

        try:

            blur_methods = ["One-Pixel-Blur", "Global Localisation Precision",
                "Individual Localisation Precision, iso", "Individual Localisation Precision"]

            if blur_method not in blur_methods:
                print(f"Blur method must be one of: {blur_methods}")
                return None

            locs = self.get_locs(symmetry=symmetry, transformed=True)

            if locs is None:
                return None

            if len(locs) == 0:
                return None

            polygon = self.data[0].cell_polygon
            polygon_coords = np.array(polygon.exterior.coords)

            locs = pd.DataFrame(locs)

            picasso_columns = ["frame",
                               "y", "x",
                               "photons", "sx", "sy", "bg",
                               "lpx", "lpy",
                               "ellipticity", "net_gradient",
                               "group", "iterations", ]

            column_filter = [col for col in picasso_columns if col in locs.columns]

            locs = locs[column_filter]
            locs = locs.to_records(index=False)

            xmin, xmax = polygon_coords[:, 0].min(), polygon_coords[:, 0].max()
            ymin, ymax = polygon_coords[:, 1].min(), polygon_coords[:, 1].max()

            h,w = int(ymax-ymin)+3, int(xmax-xmin)+3

            viewport = [(float(0), float(0)), (float(h), float(w))]

            if blur_method == "One-Pixel-Blur":
                blur_method = "smooth"
            elif blur_method == "Global Localisation Precision":
                blur_method = "convolve"
            elif blur_method == "Individual Localisation Precision, iso":
                blur_method = "gaussian_iso"
            elif blur_method == "Individual Localisation Precision":
                blur_method = "gaussian"
            else:
                blur_method = None

            print(f"Rendering {len(locs)} localisations")

            n_rendered_locs, image = render(locs,
                viewport=viewport,
                blur_method=blur_method,
                min_blur_width=min_blur_width,
                oversampling=oversampling, ang=0, )

            cmap = self.get_custom_cmap(colour=cmap)

            plt.rcParams["axes.grid"] = False
            fig, ax = plt.subplots()
            ax.imshow(image, cmap=cmap)
            if draw_outline:
                polygon_coords = polygon_coords * oversampling
                ax.plot(*polygon_coords.T, color='white')
            ax.axis('off')

            buf = BytesIO()
            plt.savefig(buf, format='png',
                bbox_inches='tight',pad_inches=0,
                facecolor='black', dpi=dpi)
            buf.seek(0)
            image = plt.imread(buf)
            plt.close(fig)

            if show:
                plt.imshow(image)
                plt.axis('off')
                plt.show()

            if save:
                try:
                    if path is None:
                        print("Save path must be specified")
                        return None
                    plt.imsave(path, image)
                    print(f"Saved render to {path}")
                except:
                    print(traceback.format_exc())
                    return None

            plt.close()

        except:
            print(traceback.format_exc())
            return None

        return image

    def plot_cells(self,xlim=None,ylim=None,locs=True):

        try:

            for cell in copy.deepcopy(self.data):

                polygon = cell.cell_polygon
                polygon_coords = np.array(polygon.exterior.coords)

                plt.plot(*polygon_coords.T, color="black")

                if locs:
                    if cell.locs is None:
                        continue
                    if len(cell.locs) == 0:
                        continue
                    plt.scatter(cell.locs["x"], cell.locs["y"], color="blue")

            if xlim is not None:
                plt.xlim(xlim)
            if ylim is not None:
                plt.ylim(ylim)

            plt.show()


        except:
            print(traceback.format_exc())
            pass
























import tifffile 
import numpy as np
import cv2
import os
import traceback
import matplotlib.pyplot as plt
from skimage.morphology import skeletonize, medial_axis, binary_opening, disk
from scipy.interpolate import CubicSpline, BSpline
from skimage.draw import polygon
from shapely.geometry import LineString, Point, LinearRing, Polygon
from shapely.affinity import rotate
import matplotlib.path as mpltPath
from scipy.spatial.distance import cdist
from scipy.spatial import Voronoi
import matplotlib.pyplot as plt
from scipy.optimize import minimize
import copy
from scipy.interpolate import interp1d
import math
from scipy.interpolate import splrep, splev
import numpy as np
from shapely.geometry import Polygon
from scipy.spatial import Voronoi
import matplotlib.pyplot as plt
import shapely
from scipy.spatial.distance import directed_hausdorff
import shapely
from scipy.spatial import distance
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed
from multiprocessing import Manager, Event
from functools import partial
import warnings

from bactfit.utils import (resize_line, moving_average, rotate_polygon, rotate_linestring, fit_poly)

class bactfit(object):

    def __init__(self,
            cell = None,
            celllist = None,
            refine_fit = True,
            min_radius = -1,
            max_radius = -1,
            fit_mode = "directed_hausdorff",
            progress_callback=None,
            parallel=True,
            silence_tqdm=False,
            max_workers=None,
            max_error = -1,
            **kwargs):

        self.cell = cell
        self.celllist = celllist
        self.refine_fit = refine_fit
        self.min_radius = min_radius
        self.max_radius = max_radius
        self.fit_mode = fit_mode
        self.progress_callback = progress_callback
        self.parallel = parallel
        self.max_workers = max_workers
        self.silence_tqdm = silence_tqdm
        self.max_error = max_error
        self.kwargs = kwargs


    def fit(self):

        if self.cell is not None:

            fitted_cell = bactfit.fit_cell(
                self.cell,
                refine_fit=self.refine_fit,
                fit_mode=self.fit_mode,
                min_radius=self.min_radius,
                max_radius=self.max_radius,
                silence_tqdm=self.silence_tqdm,
                max_error=self.max_error,)

            return fitted_cell

        if self.celllist is not None:

            fitted_cells = bactfit.fit_celllist(
                self.celllist,
                refine_fit=self.refine_fit,
                fit_mode=self.fit_mode,
                min_radius=self.min_radius,
                max_radius=self.max_radius,
                progress_callback=self.progress_callback,
                parallel=self.parallel,
                max_workers=self.max_workers,
                silence_tqdm=self.silence_tqdm,
                max_error=self.max_error,)

            return fitted_cells


    @staticmethod
    def get_polygon_medial_axis(outline, refine=True):

        if type(outline) == LineString:
            polygon_outline = outline
            polygon = Polygon(outline.coords)
        elif type(outline) == Polygon:
            polygon = outline
            polygon_outline = LineString(polygon.exterior.coords)
        else:
            return None, None

        if len(polygon_outline.coords) < 200:
            polygon_outline = resize_line(polygon_outline, 200)
            polygon = Polygon(polygon_outline.coords)

        # Extract the exterior coordinates of the polygon
        exterior_coords = np.array(polygon.exterior.coords)

        exterior_coords = moving_average(exterior_coords, padding=10, iterations=2)

        # Compute the Voronoi diagram of the exterior coordinates
        vor = Voronoi(exterior_coords)

        # Function to check if a point is inside the polygon
        def point_in_polygon(point, polygon):
            return polygon.contains(Point(point))

        # Extract the medial axis points from the Voronoi vertices

        coords = []

        for i, region in enumerate(vor.regions):
            if -1 not in region:
                try:
                    coords.append(vor.vertices[i].tolist())
                except:
                    pass

        coords = [point for point in coords if point_in_polygon(point, polygon)]

        centroid = polygon.centroid
        cell_radius = polygon_outline.distance(centroid)

        if refine:
            coords = [p for p in coords if polygon_outline.distance(Point(p)) > cell_radius * 0.8]
            coords = np.array(coords)

        return np.array(coords), cell_radius

    @staticmethod
    def bactfit_result(cell, params, cell_polygon, poly_params,
            fit_mode = "directed_hausdorff"):

        x_min = params[0]
        x_max = params[1]
        cell_radius = params[2]
        x_offset = params[3]
        y_offset = params[4]
        angle = params[5]

        p = np.poly1d(poly_params)
        x_fitted = np.linspace(x_min, x_max, num=20)
        y_fitted = p(x_fitted)

        x_fitted += x_offset
        y_fitted += y_offset

        midline_coords = np.column_stack((x_fitted, y_fitted))
        midline = LineString(midline_coords)

        midline = rotate_linestring(midline, angle=angle)

        midline_coords = np.array(midline.coords)
        cell_poles = [midline_coords[0], midline_coords[-1]]

        fitted_polygon = midline.buffer(cell_radius)

        distance = bactfit.compute_bacfit_distance(fitted_polygon, cell_polygon, fit_mode)

        cell.cell_midline = midline
        cell.cell_polygon = fitted_polygon
        cell.cell_poles = cell_poles
        cell.polynomial_params = poly_params
        cell.fit_error = distance
        cell.cell_radius = cell_radius
        
        return cell


    @staticmethod
    def refine_function(params, cell_polygon, poly_params,
            fit_mode="directed_hausdorff"):

        """
        Objective function to minimize: the Hausdorff distance between the buffered spline and the target contour.
        """

        try:
            params = list(params)

            x_min = params[0]
            x_max = params[1]
            cell_radius = params[2]
            x_offset = params[3]
            y_offset = params[4]
            angle = params[5]

            p = np.poly1d(poly_params)
            x_fitted = np.linspace(x_min, x_max, num=10)
            y_fitted = p(x_fitted)

            x_fitted += x_offset
            y_fitted += y_offset

            midline_coords = np.column_stack((x_fitted, y_fitted))

            midline = LineString(midline_coords)
            midline = rotate_linestring(midline, angle=angle)

            midline_buffer = midline.buffer(cell_radius)

            distance = bactfit.compute_bacfit_distance(midline_buffer,
                cell_polygon, fit_mode)

        except:
            distance = np.inf

        return distance

    @staticmethod
    def compute_bacfit_distance(midline_buffer, cell_polygon,
            fit_mode = "directed_hausdorff"):

        try:

            midline_buffer_centroid = midline_buffer.centroid
            cell_polygon_centroid = cell_polygon.centroid

            if fit_mode == "hausdorff":
                # Calculate the Hausdorff distance between the buffered spline and the target contour
                distance = midline_buffer.hausdorff_distance(cell_polygon)
            elif fit_mode == "directed_hausdorff":
                # Calculate directed Hausdorff distance in both directions
                buffer_points = np.array(midline_buffer.exterior.coords)
                contour_points = np.array(cell_polygon.exterior.coords)
                dist1 = directed_hausdorff(buffer_points, contour_points)[0]
                dist2 = directed_hausdorff(contour_points, buffer_points)[0]
                distance = dist1 + dist2

            centroid_distance = midline_buffer_centroid.distance(cell_polygon_centroid)
            distance += centroid_distance**0.5

        except:
            # print(traceback.format_exc())
            distance = np.inf

        return distance

    @staticmethod
    def get_poly_coords(x, y, coefficients, margin=0, n_points=10):
        x1 = np.min(x) - margin
        x2 = np.max(x) + margin

        p = np.poly1d(coefficients)
        x_fitted = np.linspace(x1, x2, num=n_points)
        y_fitted = p(x_fitted)

        return np.column_stack((x_fitted, y_fitted))

    @staticmethod
    def register_fit_data(cell, cell_centre=[], vertical=False):

        if vertical:

            cell_polygon = cell.cell_polygon
            cell_midline = cell.cell_midline

            cell_polygon = rotate_polygon(cell_polygon, angle=-90)
            cell_midline = rotate_linestring(cell_midline, angle=-90)

            midline_coords = np.array(cell_midline.coords)
            cell_poles = [midline_coords[0], midline_coords[-1]]

            cell.cell_polygon = cell_polygon
            cell.cell_midline = cell_midline
            cell.cell_poles = cell_poles

        return cell


    @staticmethod
    def fit_cell(cell, refine_fit=True, fit_mode="directed_hausdorff",
            min_radius=-1, max_radius=-1, polygon_length = 100, max_error = -1, **kwargs):

        try:

            cell_polygon = cell.cell_polygon
            cell_outline = LineString(cell_polygon.exterior.coords)
            vertical = cell.vertical

            if len(cell_outline.coords) < polygon_length:
                cell_outline = resize_line(cell_outline, polygon_length)
                cell_polygon = Polygon(cell_outline.coords)

            if vertical:
                cell_polygon = rotate_polygon(cell_polygon)

            medial_axis_coords, radius = bactfit.get_polygon_medial_axis(cell_polygon)

            if min_radius > 0:
                radius = max(radius, min_radius)
            if max_radius > 0:
                radius = min(radius, max_radius)

            medial_axis_fit, poly_params = fit_poly(medial_axis_coords,
                degree=[1, 2, 3], maxiter=100, minimise_curvature=False)

            x_min = np.min(medial_axis_fit[:, 0])
            x_max = np.max(medial_axis_fit[:, 0])
            x_offset = 0
            y_offset = 0
            rotation = 0
            params = [x_min, x_max, radius, x_offset, y_offset, rotation]

            if refine_fit:

                warnings.filterwarnings("ignore", category=RuntimeWarning)

                bounds = [(None, None),  # x_min
                          (None, None),  # x_max
                          (None, None),  # radius
                          (None, None),  # x_offset
                          (None, None),  # y_offset
                          (None, None)]  # rotation

                if min_radius > 0:
                    bounds[2] = (min_radius, None)
                if max_radius > 0:
                    bounds[2] = (None, max_radius)

                result = minimize(bactfit.refine_function, params,
                    args=(cell_polygon, poly_params, fit_mode),
                    tol=1e-1, options={'maxiter': 500}, bounds=bounds)

                params = result.x

            cell = bactfit.bactfit_result(cell, params, cell_polygon,
                poly_params, fit_mode)

            cell = bactfit.register_fit_data(cell, vertical=vertical)

            if max_error > 0 and hasattr(cell, "fit_error"):
                if cell.fit_error > max_error:
                    cell = None

        except:
            pass

        return cell

    @staticmethod
    def fit_celllist(celllist, refine_fit=True, fit_mode="directed_hausdorff",
            min_radius=-1, max_radius=-1, progress_callback=None, parallel=True,
            max_workers=None, silence_tqdm=False, max_error=-1, **kwargs):

        num_cells = len(celllist.data)

        if parallel:

            if max_workers == None:
                max_workers = os.cpu_count()

            with ProcessPoolExecutor(max_workers=max_workers) as executor:

                futures = {executor.submit(bactfit.fit_cell, cell, refine_fit=refine_fit,
                    fit_mode=fit_mode, min_radius=min_radius, max_radius=max_radius, max_error=max_error): cell
                    for cell in celllist.data}

                completed = 0
                for future in tqdm(as_completed(futures), total=num_cells,
                        desc="Optimising Cells", disable=silence_tqdm):

                    cell = future.result()
                    if cell is not None:
                        if cell.fit_error is not None:
                            cell_index  = cell.cell_index
                            celllist.data[cell_index] = cell

                    completed += 1
                    if progress_callback is not None:
                        progress = (completed / num_cells) * 100
                        progress_callback.emit(progress)

        else:

            iter = 0

            for cell in tqdm(celllist, total=num_cells,
                    desc="Optimising Cells", disable=silence_tqdm):

                cell = bactfit.fit_cell(cell, refine_fit=refine_fit,
                    fit_mode=fit_mode, min_radius=min_radius,
                    max_radius=max_radius)

                if cell.fit_error is not None:
                    cell_index = cell.cell_index
                    celllist.data[cell_index] = cell

                if progress_callback is not None:
                    progress = ((iter + 1) / num_cells)*100
                    progress_callback.emit(progress)

                iter += 1

        #remove cells with no fit error
        celllist.data = [cell for cell in celllist.data if cell.fit_error is not None]
        celllist.assign_cell_indices(reindex=True)

        num_fitted_cells = len(celllist.data)
        print(f"Fit completed for {num_fitted_cells}/{num_cells} cells")

        return celllist



import matplotlib.pyplot as plt
from glob2 import glob
import pandas as pd
import tifffile
import numpy as np
import os
import bactfit
from glob2 import glob
from bactfit.cell import CellList, Cell, ModelCell
from bactfit.preprocess import data_to_cells, mask_to_cells
import pickle
from bactfit.fileIO import save, load


if __name__ == '__main__':

    mask = tifffile.imread("mask2.tif") # Load mask stack
    locs = pd.read_csv("localisations2.csv") # Load picasso localisations

    celllist = mask_to_cells(mask)  # Create celllist from binary mask
    celllist.add_localisations(locs)  # Add localisations to celllist






import h5py
import numpy as np
from shapely.geometry import Polygon, LineString
from bactfit.cell import Cell, CellList
import pickle
import os
import traceback

def save(file_path, cell_data):

    try:

        file_path = os.path.abspath(file_path)
        file_path = os.path.normpath(file_path)
        file_dir = os.path.dirname(file_path)

        if not os.path.exists(file_dir):
            os.makedirs(file_dir)

        file, ext = os.path.splitext(file_path)
        file_path = file + ".h5"

        if cell_data.type == 'Cell':

            with h5py.File(file_path, 'w') as f:
                if hasattr(cell_data, 'cell_index'):
                    save_name = f"cell_{cell_data.cell_index}"
                else:
                    save_name = "cell_0"

                cell_group = f.create_group(save_name)
                write_cell(cell_group, cell_data)

            print(f"Saved cell to {file_path}")

        elif cell_data.type == 'CellList':

            with h5py.File(file_path, 'w') as f:
                n_cells = len(cell_data.data)

                for cell_index, cell in enumerate(cell_data.data):
                    file_index = str(cell_index).zfill(int(np.ceil(np.log10(n_cells))))
                    save_name = f"cell_{file_index}"
                    cell_group = f.create_group(save_name)
                    write_cell(cell_group, cell)

            print(f"Saved {n_cells} cell(s) to {file_path}")

        else:
            print(f"Invalid data type: type{cell_data}")

    except:
        print(traceback.format_exc())
        print("Error saving cell data")



def write_cell(cell_group, cell):

    attr_grp = cell_group.create_group('attributes')

    attribute_list = ["name","cell_radius", "cell_length",
                      "cell_midline", "cell_polygon",
                      "cell_poles", "polynomial_params",
                      "fit_error", "pixel_size", "locs",
                      "cell_index", "vertical",'crop_bounds',
                      'cell_centre', 'bbox', 'height', 'width',
                      'transformed', "bactfit_version","type"]

    for attr in attribute_list:
        if hasattr(cell, attr):
            attr_data = cell.__getattribute__(attr)
            if attr_data is not None:
                try:
                    if isinstance(attr_data, (Polygon, LineString,np.ndarray, np.recarray)):
                        # Serialize geometric objects
                        attr_data = pickle.dumps(attr_data)
                    elif not isinstance(attr_data, (int, float, str)):
                        # Serialize other complex objects
                        attr_data = pickle.dumps(attr_data)

                    if isinstance(attr_data, bytes):
                        attr_grp.create_dataset(attr, data=np.void(attr_data))
                    else:
                        attr_grp.create_dataset(attr, data=attr_data)
                except Exception as e:
                    print(f'Could not save attribute {attr}: {e}')

    data_grp = cell_group.create_group('data')
    for channel, image in cell.data.items():
        if type(image) == np.ndarray:
            grp = data_grp.create_group(channel)
            grp.create_dataset(channel, data=image)
            grp.attrs.create('dclass', np.string_(image.dtype))

def load(file_path):

    cell_list = []

    if type(file_path) is str:
        file_path = [file_path]

    for path in file_path:

        with h5py.File(path, 'r') as f:
            cells = [load_cell(f[key]) for key in f.keys()]
            cell_list.extend(cells)

    if len(cell_list) == 1:
        print(f'Loaded cell')
        return cell_list[0]
    else:
        print(f'Loaded {len(cell_list)} cells')
        return CellList(cell_list)

def load_cell(cell_group):

    attr_grp = cell_group['attributes']


    attr_dict = {}

    for attr in list(attr_grp.keys()):
        attr_data = attr_grp[attr][()]

        if isinstance(attr_data, np.void):
            # Deserialize byte data
            attr_data = pickle.loads(bytes(attr_data))
        elif isinstance(attr_data, bytes):
            # Decode bytes to string
            attr_data = attr_data.decode('utf-8')

        attr_dict[attr] = attr_data
        
    cell = Cell(attr_dict)
    
    data_grp = cell_group['data']
    
    data_dict = {}
    
    for channel in list(data_grp.keys()):
        try:
            grp = data_grp[channel]
            data_arr = grp[channel]
            dclass = grp.attrs.get('dclass').decode('UTF-8')
            image = np.array(data_arr, dtype=dclass)
            
            if channel not in data_dict.keys():
                data_dict[channel] = image
            
        except:
            print(traceback.format_exc())
            
    cell.data = data_dict

    return cell







from bactfit.cell import Cell, CellList
from bactfit.utils import resize_line, rotate_linestring, fit_poly, get_vertical, moving_average, get_polygon_midline
from shapely.geometry import Polygon, Point, LineString
from scipy.spatial import Voronoi
import numpy as np
import cv2
import matplotlib.pyplot as plt


def data_to_cells(segmentation_list, locs = None):

    cell_list = []

    for seg in segmentation_list:

        cell_images = {}

        if seg.shape[1] == 2:
            frame_index = -1

            if len(seg) < 3:
                continue

            cell_polygon = Polygon(seg)

        if seg.shape[1] == 3:
            frame_index = seg[0, 0]

            seg = seg[1:]

            cell_polygon = Polygon(seg)

        centroid = cell_polygon.centroid
        cell_centre = [centroid.x, centroid.y]

        minx, miny, maxx, maxy = cell_polygon.bounds

        bbox = [minx, miny, maxx, maxy]

        h = maxy - miny
        w = maxx - minx

        if h > w:
            vertical = True
        else:
            vertical = False

        cell_data = {
            "cell_polygon": cell_polygon,
            "cell_centre": cell_centre,
            "bbox": bbox,
            "height": h,
            "width": w,
            "vertical": vertical,
            "frame_index": frame_index
        }

        cell = Cell(cell_data)

        cell_list.append(cell)

    if len(cell_list):
        cell_list = CellList(cell_list)

    return cell_list



def check_edge(bbox, mask_shape,
        buffer = 1):

    x1, y1, x2, y2 = bbox

    if x1 < buffer:
        edge = True
    elif y1 < buffer:
        edge = True
    elif x2 > mask_shape[1] - buffer:
        edge = True
    elif y2 > mask_shape[0] - buffer:
        edge = True
    else:
        edge = False

    return edge

def mask_to_cells(masks, images=None, locs=None, buffer = 1, flipxy = False, verbose = True):

    mask_list = None

    if isinstance(masks, np.ndarray):
        if len(masks.shape) == 3:
            mask_list = [mask for mask in masks]
        else:
            mask_list = [masks]
    if type(masks) == list:
        mask_list = masks

    if mask_list:

        cell_list = []

        for frame_index, mask in enumerate(mask_list):

            mask_ids = np.unique(mask)

            for mask_id in mask_ids:

                if mask_id == 0:
                    continue

                cell_mask = np.zeros_like(mask,dtype=np.uint8)
                cell_mask[mask == mask_id] = 255
                cell_mask = cell_mask.astype(np.uint8)

                contours, _ = cv2.findContours(cell_mask,
                    cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)

                if len(contours) > 0:

                    contour = contours[0]
                    contour = contour.squeeze()

                    if len(contour) < 3:
                        continue

                    cell_polygon = Polygon(contour)

                    centroid = cell_polygon.centroid
                    cell_centre = [centroid.x, centroid.y]

                    bbox = cell_polygon.bounds

                    x1, y1, x2, y2 = bbox

                    edge = check_edge(bbox, cell_mask.shape, buffer)

                    if edge == True:
                        continue

                    h = y2 - y1
                    w = x2 - x1

                    if h > w:
                        vertical = True
                    else:
                        vertical = False

                    cell_data = {
                        "cell_polygon": cell_polygon,
                        "cell_centre": cell_centre,
                        "bbox": bbox,
                        "height": h,
                        "width": w,
                        "vertical": vertical,
                        "frame_index": frame_index
                    }

                    cell = Cell(cell_data)

                    cell_list.append(cell)
            
    if len(cell_list):
        celllist = CellList(cell_list)

        if verbose:
            print(f"Added {len(cell_list)} cells to CellList")

    return celllist



import tifffile
from glob2 import glob
import numpy as np
import pandas as pd

data_dir = r"/example_data"

# mask_files = glob(data_dir + "/*.tif")[:5]
#
# masks = []
#
# for path in mask_files:
#
#     mask = tifffile.imread(path)
#
#     masks.append(mask)
#
# masks = np.stack(masks)
#
# tifffile.imwrite("../../example_data/mask_stack.tif", masks)


csv_files = glob(data_dir + "/*.csv")

df = pd.read_csv(csv_files[0])

df = df[df["frame"] <=5]

df.to_csv(csv_files[0], index=False)

with open('celllist.pickle', 'wb') as handle:
    pickle.dump(celllist, handle, protocol=pickle.HIGHEST_PROTOCOL)

with open('celllist.pickle', 'rb') as handle:
    celllist = pickle.load(handle)


import tifffile
import numpy as np
import cv2
import os
import traceback
import matplotlib.pyplot as plt
from skimage.morphology import skeletonize, medial_axis, binary_opening, disk
from scipy.interpolate import CubicSpline, BSpline
from skimage.draw import polygon
from shapely.geometry import LineString, Point, LinearRing, Polygon
from shapely.affinity import rotate
import matplotlib.path as mpltPath
from scipy.spatial.distance import cdist
from scipy.spatial import Voronoi
import matplotlib.pyplot as plt
from scipy.optimize import minimize
import copy
from scipy.interpolate import interp1d
import math
from scipy.interpolate import splrep, splev
import numpy as np
from shapely.geometry import Polygon
from scipy.spatial import Voronoi
import matplotlib.pyplot as plt
import shapely
from scipy.spatial.distance import directed_hausdorff
import shapely
from scipy.spatial import distance
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed
from multiprocessing import Manager, Event
from functools import partial
import warnings


def optimise_cell(cell, refine_fit=True, fit_mode="directed_hausdorff", min_radius=-1, max_radius=-1):
    cell.optimise(refine_fit=refine_fit, fit_mode=fit_mode, min_radius=min_radius, max_radius=max_radius)

    return cell

def resize_line(line, length):
    distances = np.linspace(0, line.length, length)
    line = LineString([line.interpolate(distance) for distance in distances])

    return line

def moving_average(line, padding=5, iterations=1):
    x, y = line[:, 0], line[:, 1]

    x = np.concatenate((x[-padding:], x, x[:padding]))
    y = np.concatenate((y[-padding:], y, y[:padding]))

    for i in range(iterations):
        y = np.convolve(y, np.ones(padding), 'same') / padding
        x = np.convolve(x, np.ones(padding), 'same') / padding

        x = np.array(x)
        y = np.array(y)

    x = x[padding:-padding]
    y = y[padding:-padding]

    line = np.stack([x, y]).T

    return line

def rotate_polygon(polygon, angle=90):
    origin = polygon.centroid.coords[0]
    polygon = shapely.affinity.rotate(polygon, angle=angle, origin=origin)

    return polygon

def rotate_linestring(linestring, angle=90):
    centroid = linestring.centroid
    origin = centroid.coords[0]

    linestring = shapely.affinity.rotate(linestring, angle=angle, origin=origin)

    return linestring

def get_vertical(polygon):

    minx, miny, maxx, maxy = polygon.bounds

    h = maxy - miny
    w = maxx - minx

    if h > w:
        vertical = True
    else:
        vertical = False

    return vertical


def fit_poly(coords, degree=2, constrained=True, constraining_points=[],
        minimise_curvature=True, curvature_weight = 0.1, degree_penalty=0.01, maxiter=50):

    warnings.filterwarnings("ignore", category=np.RankWarning)

    def polynomial_fit(params, x):
        # Reverse the parameters to match np.polyfit order
        params = params[::-1]
        return sum(p * x ** i for i, p in enumerate(params))

    def objective_function(params, x, y, minimise_curvature=True):
        fit_error = np.sum((polynomial_fit(params, x) - y) ** 2)

        if minimise_curvature:
            curvature_penalty = np.sum(np.diff(params, n=2) ** 2)
            fit_error = fit_error + (curvature_penalty*curvature_weight)

        fit_error = fit_error + (degree_penalty * len(params))

        return fit_error

    def constraint_function(params, x_val, y_val):
        return polynomial_fit(params, x_val) - y_val

    def get_coords(x, y, coefficients, margin=0, n_points=10):
        x1 = np.min(x) - margin
        x2 = np.max(x) + margin

        p = np.poly1d(coefficients)
        x_fitted = np.linspace(x1, x2, num=n_points)
        y_fitted = p(x_fitted)

        return np.column_stack((x_fitted, y_fitted))

    try:

        x = coords[:, 0]
        y = coords[:, 1]
        constraints = []

        param_list = []
        error_list = []
        success_list = []

        if constrained and len(constraining_points) > 0:
            for point in constraining_points:
                if len(point) == 2:
                    constraints.append({'type': 'eq', 'fun': constraint_function, 'args': point})

        if type(degree) != list:
            degree = [degree]

        for deg in degree:
            params = np.polyfit(x, y, deg)

            result = minimize(objective_function, params, args=(x, y, minimise_curvature),
                constraints=constraints, tol=1e-6, options={'maxiter': maxiter})

            param_list.append(result.x)
            error_list.append(result.fun)
            success_list.append(result.success)

        min_error_index = error_list.index(min(error_list))

        best_params = param_list[min_error_index]

        fitted_poly = get_coords(x, y, best_params)

    except:
        return None, None

    return fitted_poly, list(best_params)


def manual_fit(cell_coords, midline_coords, radius = None):

    cell_polygon = Polygon(cell_coords)
    cell_outline = LineString(cell_coords)

    vertical = get_vertical(cell_polygon)

    n_medial_points = len(midline_coords)

    if vertical:
        cell_polygon = rotate_polygon(cell_polygon)
        cell_coords = np.array(cell_polygon.exterior.coords)
        cell_midline = LineString(midline_coords)
        cell_midline = rotate_linestring(cell_midline)
        midline_coords = np.array(cell_midline.coords)

    constraining_points = [midline_coords[0].tolist(),
                           midline_coords[-1].tolist()]

    medial_axis_fit, poly_params = fit_poly(midline_coords,
        degree=[1, 2, 3], maxiter=100, minimise_curvature=False,
        constraining_points=constraining_points, constrained=False)

    if radius is None:
        centroid = cell_polygon.centroid
        cell_radius = cell_outline.distance(centroid)
    else:
        cell_radius = radius

    cell_midline = LineString(medial_axis_fit)
    cell_polygon = cell_midline.buffer(cell_radius)

    if vertical:
        cell_polygon = rotate_polygon(cell_polygon, angle=-90)
        cell_midline = rotate_linestring(cell_midline, angle=-90)

    polygon_coords = np.array(cell_polygon.exterior.coords)
    polygon_coords = polygon_coords[:-1]

    cell_midline = resize_line(cell_midline, n_medial_points)
    midline_coords = np.array(cell_midline.coords)

    return polygon_coords, midline_coords, poly_params, cell_radius

def resize_polygon(self, polygon, n_points):

    outline = np.array(polygon.exterior.coords)
    outline = outline[1:]

    outline = LineString(outline)

    distances = np.linspace(0, outline.length, n_points)
    outline = LineString([outline.interpolate(distance) for distance in distances])

    outline = outline.coords

    polygon = Polygon(outline)

    return polygon




def get_polygon_midline(outline, refine=True):
    
    try:

        if type(outline) == LineString:
            polygon_outline = outline
            polygon = Polygon(outline.coords)
        elif type(outline) == Polygon:
            polygon = outline
            polygon_outline = LineString(polygon.exterior.coords)
        else:
            return None, None
    
        if len(polygon_outline.coords) < 200:
            polygon_outline = resize_line(polygon_outline, 200)
            polygon = Polygon(polygon_outline.coords)
    
        # Extract the exterior coordinates of the polygon
        exterior_coords = np.array(polygon.exterior.coords)
    
        exterior_coords = moving_average(exterior_coords, padding=10, iterations=2)
    
        # Compute the Voronoi diagram of the exterior coordinates
        vor = Voronoi(exterior_coords)
    
        # Function to check if a point is inside the polygon
        def point_in_polygon(point, polygon):
            return polygon.contains(Point(point))
    
        # Extract the medial axis points from the Voronoi vertices
    
        coords = []
    
        for i, region in enumerate(vor.regions):
            if -1 not in region:
                try:
                    coords.append(vor.vertices[i].tolist())
                except:
                    pass
    
        coords = [point for point in coords if point_in_polygon(point, polygon)]
    
        centroid = polygon.centroid
        cell_radius = polygon_outline.distance(centroid)
    
        if refine:
            coords = [p for p in coords if polygon_outline.distance(Point(p)) > cell_radius * 0.8]
            coords = np.array(coords)
            
        midline_coords, poly_params = fit_poly(coords,
            degree=[1, 2, 3], maxiter=100, minimise_curvature=False)
        
        midline = LineString(midline_coords)
    
    except:
        midline = None
        buffer = None

    return midline, cell_radius

